# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom. All rights reserved.

__all__ = ['base', 'graph', 'ontology']

from knowledge.services import base
from knowledge.services import graph
from knowledge.services import ontology
