# -*- coding: utf-8 -*-
# Copyright © 2021 Wacom. All rights reserved.

__all__ = ['access', 'entity', 'ontology']

from knowledge.base import access
from knowledge.base import entity
from knowledge.base import ontology
